body {
    background-color: #333;
    font-family: sans-serif;
    font-size: 30px;
}

button {
    font-family: sans-serif;
    font-size: 30px;
    width: 32%;

    background-color: #0085cf;
    background-image: linear-gradient(to bottom, #00a8e1 0%, #0085cf 100%);
    background-repeat: repeat-x;

    border: 2px solid #ccc;
    color: #fff;

    text-transform: uppercase;

    -webkit-box-shadow: 2px 2px 10px 0px rgba(0,0,0,0.5);
    -moz-box-shadow: 2px 2px 10px 0px rgba(0,0,0,0.5);
    box-shadow: 2px 2px 10px 0px rgba(0,0,0,0.5);
}

button:hover {
    background-color: #006ba6;
    background-image: none;
    -webkit-box-shadow: none;
    -moz-box-shadow: none;
    box-shadow: none;
}

hr {
    border: none;
    background-color: #eee;
    height: 10px;
}

.menu {
    padding: 10px;
    margin-bottom: 10px;
}

.content {
    background-color: #eee;
    border: 1px solid #ccc;
    padding: 10px;

    -webkit-box-shadow: 2px 2px 10px 0px rgba(0,0,0,0.5);
    -moz-box-shadow: 2px 2px 10px 0px rgba(0,0,0,0.5);
    box-shadow: 2px 2px 10px 0px rgba(0,0,0,0.5);
}

.content .message {
    padding: 20px;
    margin-top: 10px;
    background-color: #fff;
    border: 1px solid #ccc;
    font-size: 40px;
    font-weight: bold;
}

.wrapper {
    position: absolute;
    height: 230px;
    width: 640px;
    left: 50%;
    top: 50%;
    -webkit-transform: translateX(-50%) translatey(-50%);
}

.error {
    color: #a21e22;
}

.service {
    background-color: #555;
    color: #ccc;
    font-size: 20px;
    position: fixed;
    bottom: 0px;
    left: 0px;
    right: 0px;
    padding: 5px;
}